# Frozen-string-literal: true
# rubocop:disable Layout/EmptyLinesAroundBlockBody
# Copyright: 2018 - MIT License
# Encoding: utf-8

require "msgpack"
require "pathutil"
require "json"

module MimeType
  module_function
  JSON = Pathutil.new(__dir__).join("mimetype.json")
  KEYS = %i(content_type content_types extension extensions).freeze
  STRING_KEYS = %w(content_type content_types extension extensions).freeze
  PACK = Pathutil.new(__dir__).join("mimetype.msgpack")

  # --
  # @return [Struct] -> MimeType::T
  # Creates a new struct for wrapping out MimeTypes
  # @note this is universal
  # --
  T = Struct.new(*KEYS) do

    # --
    # Converts MimeType::T to pack
    # @note converts to array then to pack
    # @return [String] the pack
    # --
    def to_msgpack_ext
      to_a.to_msgpack
    end

    # --
    class << self

      # --
      # @return [MimeType::T] struct
      # @note takes a packed array, unpacks to MimeType::T
      # @param o [String] the pack
      # --
      def from_msgpack_ext(o)
        new(*MessagePack.unpack(o))
      end
    end
  end

  # --
  # @return [Hash<String,Hash<Symbol,String>>]
  # @note content_types is mostly used by us, when they are updated
  # Convert a mime file (read: `/etc/mime.types`) into JSON
  # @param file [String, File] the path
  # --
  def parse_etc_mimetypes(file = "mime.types")
    out = Pathutil.new(file).each_line.with_object({}) do |v, h|
      v = v.split(%r!\s+!)
      next if v.size < 2 || v[0] =~ %r!^#!
      k, v = v[0], add_dots(v[1..-1])

      h[k] = {
        content_type: k,
        content_types: [],
        extension: v[0],
        extensions: v,
      }
    end

    Hash[out.sort_by do |k, _|
      k
    end]
  end

  # --
  # @return [Hash<Symbol,Hash<String,Struct>>]
  # @note makes it a double hash for fast parsing
  # Transform the mimetype.json file
  # --
  def transform_mimetype_json
    out = {}

    out[:c] = JSON.read_json
    out[:c].each do |k, v|
      out[:c][k] = T.new(*v.values_at(*STRING_KEYS))
    end

    add_t(out)
    add_e(out)
    out
  end

  # --
  # @return [Hash<Symbol,Hash<String,Struct>>]
  # List of mime_types
  # --
  def to_h
    return @to_h if @to_h
    msgp = Pathutil.new(__dir__).join("mimetype.msgpack").read
    @to_h = MessagePack.unpack(msgp)
  end

  # --
  # @return [MimeType::MimeType]
  # @param path [String, Pathutil, Pathname, File] the path
  # Find by path
  # --
  def lookup_by_filename(v)
    v = File.extname(v)
    to_h[:e][v]
  end

  # --
  # Find by file extension
  # @param ext [String] the extension in "." format
  # @return [Struct]
  # --
  def lookup_by_extension(v)
    v = v.extension if v.is_a?(T)
    to_h[:e][v]
  end

  # --
  # Find by content type
  # @param content_type [String] the content type
  # @return [Struct]
  # --
  def lookup_by_content_type(v)
    v = v.content_type if v.is_a?(T)
    to_h[:c][v]
  end

  # --
  def add_dots(array)
    array.map do |e|
      ".#{e.gsub(%r!^\.!, '')}"
    end
  end

  # --
  def add_t(out)
    out[:c].keys.each do |k|
      next unless out[:c][k].content_types.size > 1
      out[:c][k].content_types.each do |nk|
        out[:c][nk] = out[:c][k]
      end
    end
  end

  # --
  def add_e(out)
    out[:e] = out[:c].each_with_object({}) do |(_, v), h|
      v.extensions.each do |e|
        h[e] = v
      end
    end
  end
end

# --
MessagePack::DefaultFactory.register_type(0, Symbol)
MessagePack::DefaultFactory.register_type(1,
  MimeType::T)
